@@ -7,21 +7,15 @@
 #
 # Author: Kyle Lahnakoski (kyle@lahnakoski.com)
 #
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import unicode_literals
+from __future__ import absolute_import, division, unicode_literals
 
 import itertools
-from collections import Mapping
 from numbers import Number
 
-from mo_future import text_type
-
 from jx_base.expressions import jx_expression
 from mo_collections.unique_index import UniqueIndex
-from mo_dots import coalesce, Data, set_default, Null, listwrap
-from mo_dots import wrap
-from mo_dots.lists import FlatList
+from mo_dots import Data, FlatList, Null, coalesce, is_container, is_data, listwrap, set_default, unwrap, wrap
+from mo_future import text_type
 from mo_logs import Log
 from mo_math import MAX, MIN
 from mo_times.dates import Date
@@ -93,7 +87,7 @@ class Domain(object):
         return output
 
     def getDomain(self):
-        Log.error("Not implemented")
+        raise NotImplementedError()
 
     def verify_attributes_not_null(self, attribute_names):
         for name in attribute_names:
@@ -178,8 +172,16 @@ class DefaultDomain(Domain):
         self.map[key] = canonical
         return canonical
 
-    # def getIndexByKey(self, key):
-    #     return self.map.get(key).dataIndex;
+    def getIndexByKey(self, key):
+        canonical = self.map.get(key)
+        if canonical:
+            return canonical.dataIndex
+
+        index = len(self.partitions)
+        canonical = Data(name=key, value=key, dataIndex=index)
+        self.partitions.append(canonical)
+        self.map[key] = canonical
+        return index
 
     def getKey(self, part):
         return part.value
@@ -202,7 +204,12 @@ class SimpleSetDomain(Domain):
     DOMAIN IS A LIST OF OBJECTS, EACH WITH A value PROPERTY
     """
 
-    __slots__ = ["NULL", "partitions", "map", "order"]
+    __slots__ = [
+        "NULL",       # THE value FOR NULL
+        "partitions", # LIST OF {name, value, dataIndex} dicts
+        "map",        # MAP FROM value TO name
+        "order"       # MAP FROM value TO dataIndex
+    ]
 
     def __init__(self, **desc):
         Domain.__init__(self, **desc)
@@ -238,15 +245,18 @@ class SimpleSetDomain(Domain):
         if desc.partitions and desc.dimension.fields and len(desc.dimension.fields) > 1:
             self.key = desc.key
             self.map = UniqueIndex(keys=desc.dimension.fields)
-        elif desc.partitions and isinstance(desc.key, (list, set)):
+        elif desc.partitions and is_container(desc.key):
             # TODO: desc.key CAN BE MUCH LIKE A SELECT, WHICH UniqueIndex CAN NOT HANDLE
             self.key = desc.key
             self.map = UniqueIndex(keys=desc.key)
-        elif desc.partitions and isinstance(desc.partitions[0][desc.key], Mapping):
+        elif desc.partitions and is_data(desc.partitions[0][desc.key]):
+            # LOOKS LIKE OBJECTS
+            # sorted = desc.partitions[desc.key]
+
             self.key = desc.key
             self.map = UniqueIndex(keys=desc.key)
-            # self.key = UNION(set(d[desc.key].keys()) for d in desc.partitions)
-            # self.map = UniqueIndex(keys=self.key)
+            self.order = {p[self.key]: p.dataIndex for p in desc.partitions}
+            self.partitions = desc.partitions
         elif len(desc.partitions) == 0:
             # CREATE AN EMPTY DOMAIN
             self.key = "value"
@@ -380,11 +390,11 @@ class SetDomain(Domain):
         elif desc.partitions and desc.dimension.fields and len(desc.dimension.fields) > 1:
             self.key = desc.key
             self.map = UniqueIndex(keys=desc.dimension.fields)
-        elif desc.partitions and isinstance(desc.key, (list, set)):
+        elif desc.partitions and is_container(desc.key):
             # TODO: desc.key CAN BE MUCH LIKE A SELECT, WHICH UniqueIndex CAN NOT HANDLE
             self.key = desc.key
             self.map = UniqueIndex(keys=desc.key)
-        elif desc.partitions and isinstance(desc.partitions[0][desc.key], Mapping):
+        elif desc.partitions and is_data(desc.partitions[0][desc.key]):
             self.key = desc.key
             self.map = UniqueIndex(keys=desc.key)
             # self.key = UNION(set(d[desc.key].keys()) for d in desc.partitions)
@@ -655,7 +665,7 @@ class RangeDomain(Domain):
             if not self.key:
                 Log.error("Must have a key value")
 
-            parts = list(listwrap(self.partitions))
+            parts =listwrap(self.partitions)
             for i, p in enumerate(parts):
                 self.min = MIN([self.min, p.min])
                 self.max = MAX([self.max, p.max])
@@ -667,10 +677,10 @@ class RangeDomain(Domain):
 
             # VERIFY PARTITIONS DO NOT OVERLAP, HOLES ARE FINE
             for p, q in itertools.product(parts, parts):
-                if p is not q and p.min <= q.min and q.min < p.max:
+                if p.min <= q.min and q.min < p.max and unwrap(p) is not unwrap(q):
                     Log.error("partitions overlap!")
 
-            self.partitions = parts
+            self.partitions = wrap(parts)
             return
         elif any([self.min == None, self.max == None, self.interval == None]):
             Log.error("Can not handle missing parameter")
