@@ -5,17 +5,14 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this file,
 # You can obtain one at http://mozilla.org/MPL/2.0/.
 #
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import unicode_literals
+from __future__ import absolute_import, division, unicode_literals
 
 import re
 
-from mo_future import text_type
-
+from mo_future import is_text
 from mo_logs import Log
 
-keyword_pattern = re.compile(r"(\w|[\\.,$])+(?:\.(\w|[\\.,$])+)*")
+keyword_pattern = re.compile(r"(\w|[\\.,$-])+(?:\.(\w|[\\.,$-])+)*")
 
 
 def is_variable_name(value):
@@ -23,7 +20,7 @@ def is_variable_name(value):
         Log.warning("not expected")
         return True
 
-    if not value or not isinstance(value, text_type):
+    if not value or not is_text(value):
         return False  # _a._b
     value = value.lstrip(".")
     if not value:
